/*
 * ffsox_lame.c
 * Copyright (C) 2011, 2012 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "ffsox_rep.h"

#ifndef WIN32
  #define stricmp strcasecmp
  #define strnicmp strncasecmp
#endif

//#define FLT_LAME_GAIN

///////////////////////////////////////////////////////////////////////////////
typedef struct mp3buf {
  uint8_t *buf;
  uint8_t *wp;
  uint8_t *mp;
  ffsox_rep_t r;
} mp3buf_t;

static mp3buf_t *mp3buf_init(mp3buf_t *mp3)
{
  int size=16384;

  memset(mp3,0,sizeof *mp3);

  if (NULL==ffsox_rep_init(&mp3->r))
	goto error;

  if (NULL==(mp3->buf=malloc(size)))
    goto error;

  mp3->wp=mp3->buf;
  mp3->mp=mp3->buf+size;

  return mp3;
error:
  return NULL;
}

static mp3buf_t *mp3buf_cleanup(mp3buf_t *mp3)
{
  ffsox_rep_cleanup(&mp3->r);
  free(mp3->buf);

  return mp3;
}

static mp3buf_t *mp3buf_reset(mp3buf_t *mp3)
{
  mp3->wp=mp3->buf;

  return mp3;
}

static mp3buf_t *mp3buf_realloc(mp3buf_t *mp3, int len)
{
  int size=mp3->mp-mp3->buf;
  int offs=mp3->wp-mp3->buf;
  int touched=0;
  uint8_t *buf;

  while (size<offs+len) {
    size<<=1;
    touched=1;
  }

  if (touched) {
    if (NULL==(buf=realloc(mp3->buf,size)))
      goto error;

    mp3->buf=buf;
    mp3->wp=buf+offs;
    mp3->mp=buf+size;
  }

  return mp3;
error:
  return NULL;
}

#define LAME_ENCODE_INTERLEAVED(T,lame_encode,data) do { \
  for (;;) { \
    x=lame_encode( \
        l, \
        (T *)(data), \
        frame->nb_samples, \
        mp3->wp, \
        mp3->mp-mp3->wp); \
 \
    if (-1==x) { \
      if (NULL==mp3buf_realloc(mp3,2*(mp3->mp-mp3->buf))) \
        goto error; \
    } \
    else if (x<0) \
      goto error; \
    else \
      break; \
  } \
} while (0)

#define LAME_ENCODE_PLANAR(T,lame_encode,data) do { \
  T *pcm_l=(T *)(data); \
  T *pcm_r=pcm_l+frame->nb_samples; \
 \
  for (;;) { \
    x=lame_encode( \
        l, \
        pcm_l, \
        pcm_r, \
        frame->nb_samples, \
        mp3->wp, \
        mp3->mp-mp3->wp); \
 \
    if (-1==x) { \
      if (NULL==mp3buf_realloc(mp3,2*(mp3->mp-mp3->buf))) \
        goto error; \
    } \
    else if (x<0) \
      goto error; \
    else \
      break; \
  } \
} while (0)

#if ! defined (FLT_LAME_GAIN)
#define LAME_ENCODE_FLT(T,lame_encode,data,planar) do { \
  T *wp=(T *)(data); \
  T *mp=wp+2*frame->nb_samples; \
 \
  while (wp<mp) \
    *wp++*=gain; \
 \
  LAME_ENCODE_##planar(T,lame_encode,data); \
} while (0)
#else
#define LAME_ENCODE_FLT(T,lame_encode,data,planar) \
  LAME_ENCODE_##planar(T,lame_encode,data)
#endif

static mp3buf_t *mp3buf_enc(mp3buf_t *mp3, lame_t l, AVCodecContext *dc,
    AVFrame *frame, double gain)
{
  int len=ceil(1.25*frame->nb_samples+7200);
  int x=0;
  uint8_t *buf;

  if (NULL==mp3buf_realloc(mp3,len))
    goto error;

  switch (dc->sample_fmt) {
    case AV_SAMPLE_FMT_S16:
      LAME_ENCODE_INTERLEAVED(int16_t,lame_encode_buffer_interleaved,
          frame->data[0]);
      break;
    case AV_SAMPLE_FMT_S16P:
      LAME_ENCODE_PLANAR(int16_t,lame_encode_buffer,frame->data[0]);
      break;
    case AV_SAMPLE_FMT_S32:
      FFSOX_REP_I2P(int32_t,&mp3->r,frame,2,error);
      LAME_ENCODE_PLANAR(int32_t,lame_encode_buffer_int,mp3->r.buf);
      break;
    case AV_SAMPLE_FMT_S32P:
      LAME_ENCODE_PLANAR(int32_t,lame_encode_buffer_int,frame->data[0]);
      break;
    case AV_SAMPLE_FMT_FLT:
      LAME_ENCODE_FLT(float,lame_encode_buffer_interleaved_ieee_float,
          frame->data[0],INTERLEAVED);
      break;
    case AV_SAMPLE_FMT_FLTP:
      LAME_ENCODE_FLT(float,lame_encode_buffer_ieee_float,
          frame->data[0],PLANAR);
      break;
    case AV_SAMPLE_FMT_DBL:
      LAME_ENCODE_FLT(double,lame_encode_buffer_interleaved_ieee_double,
          frame->data[0],INTERLEAVED);
      break;
    case AV_SAMPLE_FMT_DBLP:
      LAME_ENCODE_FLT(double,lame_encode_buffer_ieee_double,
          frame->data[0],PLANAR);
      break;
    default:
      x=0;
      break;
  }

  if (0<x)
    mp3->wp+=x;

  return mp3;
error:
  return NULL;
}

static mp3buf_t *mp3buf_write(mp3buf_t *mp3, lame_t l,
    size_t CDECL (*fn)(lame_t,unsigned char*,size_t))
{
  int size,x;

  if (NULL==mp3buf_realloc(mp3,7200))
    goto error;

  while ((size=mp3->mp-mp3->wp)<(x=fn(l,mp3->wp,mp3->mp-mp3->wp))) {
    if (NULL==mp3buf_realloc(mp3,x))
      goto error;
  }
    
  if (0<x)
    mp3->wp+=x;

  return mp3;
error:
  return NULL;
}

static mp3buf_t *mp3buf_write_c(mp3buf_t *mp3, lame_t l,
    size_t CDECL (*fn)(const lame_global_flags *,unsigned char*,size_t))
{
  int size,x;

  if (NULL==mp3buf_realloc(mp3,7200))
    goto error;

  while ((size=mp3->mp-mp3->wp)<(x=fn(l,mp3->wp,mp3->mp-mp3->wp))) {
    if (NULL==mp3buf_realloc(mp3,x))
      goto error;
  }
    
  if (0<x)
    mp3->wp+=x;

  return mp3;
error:
  return NULL;
}

static mp3buf_t *mp3buf_flush(mp3buf_t *mp3, lame_t l)
{
  int size,x;

  if (NULL==mp3buf_realloc(mp3,7200))
    goto error;

  x=lame_encode_flush(
        l,
        mp3->wp,
        mp3->mp-mp3->wp);

  if (0<x)
    mp3->wp+=x;

  // write id3v1
  while ((size=mp3->mp-mp3->wp)<(x=lame_get_id3v1_tag(l,mp3->wp,
       mp3->mp-mp3->wp))) {

    if (NULL==mp3buf_realloc(mp3,x))
      goto error;
  }
    
  if (0<x)
    mp3->wp+=x;

  return mp3;
error:
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
typedef struct x {
  AVFormatContext *ic;
  AVCodecContext *dc;
  int index;
  AVFrame *frame;
  FILE *f;
  mp3buf_t mp3;
  lame_t l;
} x_t;

static x_t *x_cleanup(x_t *x)
{
  if (NULL!=x->l)
    lame_close(x->l);

  mp3buf_cleanup(&x->mp3);

  if (NULL!=x->frame)
    av_free(x->frame);

  if (NULL!=x->f)
    fclose(x->f);

  return x;
}

#if defined (WIN32)
static wchar_t* utf8ToUtf16(char const* mbstr)
{
  // additional Byte-Order-Marker
  int n=MultiByteToWideChar(CP_UTF8, 0, mbstr, -1, NULL, 0);
  wchar_t *wstr=malloc((n+1)*sizeof(wstr[0]));

  if (NULL!=wstr) {
    wstr[0] = 0xfeff; // BOM
    n=MultiByteToWideChar(CP_UTF8,0,mbstr,-1,wstr+1,n);

    if (0==n) {
      free(wstr);
      wstr=NULL;
    }
  }

  return wstr;
}

static void x_set_textinfo(x_t *x,char *key,char *value)
{
  wchar_t *wvalue=NULL;

  if (NULL==(wvalue=utf8ToUtf16(value)))
	goto cleanup;

  id3tag_set_textinfo_utf16(x->l,key,wvalue);
cleanup:
  if (NULL!=wvalue)
    free(wvalue);
}
#endif

static x_t *x_init(x_t *x, ffi_t *ffi, double gain, double quality,
    pb_str_t *path)
{
  AVDictionaryEntry *mtag=NULL;

//fprintf(stderr,">>> %.2f, %.1f\n",20.0*log10(gain),quality);
  memset(x,0,sizeof *x);
  x->ic=ffi->ic;
  x->dc=ffi->ast->dc;
  x->index=ffi->ast->index;

#if defined (WIN32)
  if (NULL==(x->f=_tfopen(PS_TCS(path),_T("wb"))))
    goto error;
#else
  if (NULL==(x->f=fopen(PS_UTF8(path),"wb")))
    goto error;
#endif

  if (NULL==(x->frame=avcodec_alloc_frame()))
    goto error;

  avcodec_get_frame_defaults(x->frame);

  if (NULL==mp3buf_init(&x->mp3))
    goto error;

  if (NULL==(x->l=lame_init()))
    goto error;

  // input
  lame_set_in_samplerate(x->l,x->dc->sample_rate);
  //lame_set_num_channels(x->l,dc->channels);
  lame_set_num_channels(x->l,2);

#if defined (FLT_LAME_GAIN)
  lame_set_scale(x->l,gain);
#else
  switch (x->dc->sample_fmt) {
    case AV_SAMPLE_FMT_FLT:
    case AV_SAMPLE_FMT_DBL:
    case AV_SAMPLE_FMT_FLTP:
    case AV_SAMPLE_FMT_DBLP:
      break;
    default:
      lame_set_scale(x->l,gain);
      break;
  }
#endif

  // encoder
  if (vbr_off==lame_get_VBR(x->l))
    lame_set_VBR(x->l,vbr_default);

  lame_set_VBR_quality(x->l,quality);

  // id3tag
  id3tag_init(x->l);
  id3tag_add_v2(x->l);

  while (NULL!=(mtag=av_dict_get(x->ic->metadata,"",mtag,
      AV_DICT_IGNORE_SUFFIX))) {
#if defined (WIN32)
    if (!stricmp("artist",mtag->key))
      x_set_textinfo(x,"TPE1",mtag->value);
    else if (!stricmp("album",mtag->key))
      x_set_textinfo(x,"TALB",mtag->value);
    else if (!stricmp("year",mtag->key)||!stricmp("date",mtag->key))
      x_set_textinfo(x,"TYER",mtag->value);
    else if (!stricmp("track",mtag->key))
      x_set_textinfo(x,"TRCK",mtag->value);
    else if (!stricmp("title",mtag->key))
      x_set_textinfo(x,"TIT2",mtag->value);
    else if (!stricmp("genre",mtag->key))
      x_set_textinfo(x,"TCON",mtag->value);
#else
    if (!stricmp("artist",mtag->key))
      id3tag_set_artist(x->l,mtag->value);
    else if (!stricmp("album",mtag->key))
      id3tag_set_album(x->l,mtag->value);
    else if (!stricmp("year",mtag->key)||!stricmp("date",mtag->key))
      id3tag_set_year(x->l,mtag->value);
    else if (!stricmp("track",mtag->key))
      id3tag_set_track(x->l,mtag->value);
    else if (!stricmp("title",mtag->key))
      id3tag_set_title(x->l,mtag->value);
    else if (!stricmp("genre",mtag->key))
      id3tag_set_genre(x->l,mtag->value);
#endif
  }

  // initialiazation done
  lame_set_write_id3tag_automatic(x->l,0);
  lame_init_params(x->l);

  return x;
error:
  return NULL;
}

static int x_write(x_t *x)
{
  mp3buf_t *mp3=&x->mp3;
  size_t size=mp3->wp-mp3->buf;

  if (size!=fwrite(mp3->buf,1,size,x->f))
    return -1;

  mp3->wp=mp3->buf;

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
// http://permalink.gmane.org/gmane.comp.audio.mp3.lame/4005
int ffsox_lame(pb_str_t *fn1, pb_str_t *fn2, int force_stereo, double gain,
    double quality)
{
  int code=-1;
  ffi_t ffi;
  AVPacket p1,p2;
  int audio_pos;
  int got_frame;
  x_t x;

  memset(&ffi,0,sizeof ffi);
  memset(&p1,0,sizeof p1);
  memset(&p2,0,sizeof p2);

//fprintf(stderr,"1.1 ...\n");
  if (NULL==ffi_init(&ffi,fn1,1,force_stereo))
	goto cleanup;

//fprintf(stderr,"1.2 ...\n");
  if (NULL==x_init(&x,&ffi,gain,quality,fn2))
    goto cleanup;

//fprintf(stderr,"1.3 ...\n");
  if (NULL==mp3buf_write(&x.mp3,x.l,lame_get_id3v2_tag))
    goto cleanup;
  else if (0!=x_write(&x))
    goto cleanup;

//fprintf(stderr,"1.4 ...\n");
  audio_pos=ftell(x.f);

//fprintf(stderr,"1.5 ...\n");
  while (0==av_read_frame(x.ic,&p1)) {
    if (p1.stream_index==x.index) {
      p2=p1;

      while (0<p2.size) {
//fprintf(stderr,"1.6 ...\n");
        int len=avcodec_decode_audio4(x.dc,x.frame,&got_frame,&p2);
//fprintf(stderr,"1.7 ...\n");

        if (len<0)
          goto cleanup;

        p2.size-=len;
        p2.data+=len;

        if (got_frame) {
//fprintf(stderr,"1.8 ...\n");
          if (NULL==mp3buf_enc(&x.mp3,x.l,x.dc,x.frame,gain))
            goto cleanup;

//fprintf(stderr,"1.9 ...\n");
          if (0!=x_write(&x))
            goto cleanup;
//fprintf(stderr,"1.10 ...\n");
        }
      }
    }

//fprintf(stderr,"1.11 ...\n");
    av_free_packet(&p1);
//fprintf(stderr,"1.12 ...\n");
  }

//fprintf(stderr,"1.13 ...\n");
  memset(&p2,0,sizeof p2);

  for (;;) {
     avcodec_decode_audio4(x.dc,x.frame,&got_frame,&p2);

	 if (!got_frame)
	   break;
     else if (NULL==mp3buf_enc(&x.mp3,x.l,x.dc,x.frame,gain))
       goto cleanup;
  }


  if (NULL==mp3buf_flush(&x.mp3,x.l))
    goto cleanup;
  else if (0!=x_write(&x))
    goto cleanup;

//fprintf(stderr,"1.14 ...\n");
  fseek(x.f,audio_pos,SEEK_SET);

//fprintf(stderr,"1.15 ...\n");
  if (NULL==mp3buf_write_c(&x.mp3,x.l,lame_get_lametag_frame))
    goto cleanup;
  else if (0!=x_write(&x))
    goto cleanup;

//fprintf(stderr,"1.16 ...\n");
  code=0;
cleanup:
//fprintf(stderr,"1.17 ...\n");
  x_cleanup(&x);
//fprintf(stderr,"1.18 ...\n");
  ffi_cleanup(&ffi);
//fprintf(stderr,"1.19 ...\n");

  return code;
}
